
from twisted.trial import unittest
from twisted.python import failure, filepath
from twisted.application import service

from axiom import iaxiom, store, item, attributes, batch, substore

class TestWorkUnit(item.Item):
    information = attributes.integer()

    def __repr__(self):
        return '<TestWorkUnit %d>' % (self.information,)


class ExtraUnit(item.Item):
    unformashun = attributes.text()



class WorkListener(item.Item):
    comply = attributes.integer(doc="""
    This exists solely to satisfy the requirement that Items have at
    least one persistent attribute.
    """)

    listener = attributes.inmemory(doc="""
    A callable which will be invoked by processItem.  This will be
    provided by the test method and will assert that the appropriate
    items are received, in the appropriate order.
    """)

    def processItem(self, item):
        self.listener(item)



class BatchTestCase(unittest.TestCase):
    def setUp(self):
        self.procType = batch.processor(TestWorkUnit)
        self.store = store.Store()
        self.scheduler = iaxiom.IScheduler(self.store)


    def testItemTypeCreation(self):
        """
        Test that processors for a different Item types can be
        created, that they are valid Item types themselves, and that
        repeated calls return the same object when appropriate.
        """
        procB = batch.processor(TestWorkUnit)
        self.assertIdentical(self.procType, procB)

        procC = batch.processor(ExtraUnit)
        self.failIfIdentical(procB, procC)
        self.failIfEqual(procB.typeName, procC.typeName)


    def testInstantiation(self):
        """
        Test that a batch processor can be instantiated and added to a
        database, and that it can be retrieved in the usual ways.
        """
        proc = self.procType(store=self.store)
        self.assertIdentical(self.store.findUnique(self.procType), proc)


    def testListenerlessProcessor(self):
        """
        Test that a batch processor can be stepped even if it has no
        listeners, and that it correctly reports it has no work to do.
        """
        proc = self.procType(store=self.store)
        self.failIf(proc.step(), "expected no more work to be reported, some was")

        TestWorkUnit(store=self.store, information=0)
        self.failIf(proc.step(), "expected no more work to be reported, some was")


    def testListeners(self):
        """
        Test that items can register or unregister their interest in a
        processor's batch of items.
        """
        proc = self.procType(store=self.store)
        listenerA = WorkListener(store=self.store)
        listenerB = WorkListener(store=self.store)

        self.assertEquals(list(proc.getReliableListeners()), [])

        proc.addReliableListener(listenerA)
        self.assertEquals(list(proc.getReliableListeners()), [listenerA])

        proc.addReliableListener(listenerB)
        expected = [listenerA, listenerB]
        listeners = list(proc.getReliableListeners())
        self.assertEquals(sorted(expected), sorted(listeners))

        proc.removeReliableListener(listenerA)
        self.assertEquals(list(proc.getReliableListeners()), [listenerB])

        proc.removeReliableListener(listenerB)
        self.assertEquals(list(proc.getReliableListeners()), [])


    def testBasicProgress(self):
        """
        Test that when a processor is created and given a chance to
        run, it completes some work.
        """
        processedItems = []
        def listener(item):
            processedItems.append(item.information)

        proc = self.procType(store=self.store)
        listener = WorkListener(store=self.store, listener=listener)

        proc.addReliableListener(listener)

        self.assertEquals(processedItems, [])

        self.failIf(proc.step(), "expected no work to be reported, some was")

        self.assertEquals(processedItems, [])

        for i in range(3):
            TestWorkUnit(store=self.store, information=i)
            ExtraUnit(store=self.store, unformashun=unicode(-i))

        self.failUnless(proc.step(), "expected more work to be reported, none was")
        self.assertEquals(processedItems, [0])

        self.failUnless(proc.step(), "expected more work to be reported, none was")
        self.assertEquals(processedItems, [0, 1])

        self.failIf(proc.step(), "expected no more work to be reported, some was")
        self.assertEquals(processedItems, [0, 1, 2])

        self.failIf(proc.step(), "expected no more work to be reported, some was")
        self.assertEquals(processedItems, [0, 1, 2])


    def testProgressAgainstExisting(self):
        """
        Test that when a processor is created when work units exist
        already, it works backwards to notify its listener of all
        those existing work units.  Also test that work units created
        after the processor are also handled.
        """
        processedItems = []
        def listener(item):
            processedItems.append(item.information)

        proc = self.procType(store=self.store)
        listener = WorkListener(store=self.store, listener=listener)

        for i in range(3):
            TestWorkUnit(store=self.store, information=i)

        proc.addReliableListener(listener)

        self.assertEquals(processedItems, [])

        self.failUnless(proc.step(), "expected more work to be reported, none was")
        self.assertEquals(processedItems, [2])

        self.failUnless(proc.step(), "expected more work to be reported, none was")
        self.assertEquals(processedItems, [2, 1])

        self.failIf(proc.step(), "expected no more work to be reported, some was")
        self.assertEquals(processedItems, [2, 1, 0])

        self.failIf(proc.step(), "expected no more work to be reported, some was")
        self.assertEquals(processedItems, [2, 1, 0])

        for i in range(3, 6):
            TestWorkUnit(store=self.store, information=i)

        self.failUnless(proc.step(), "expected more work to be reported, none was")
        self.assertEquals(processedItems, [2, 1, 0, 3])

        self.failUnless(proc.step(), "expected more work to be reported, none was")
        self.assertEquals(processedItems, [2, 1, 0, 3, 4])

        self.failIf(proc.step(), "expected no more work to be reported, some was")
        self.assertEquals(processedItems, [2, 1, 0, 3, 4, 5])

        self.failIf(proc.step(), "expected no more work to be reported, some was")
        self.assertEquals(processedItems, [2, 1, 0, 3, 4, 5])


    def testBrokenListener(self):
        """
        Test that if a listener's processItem method raises an
        exception, processing continues beyond that item and that an
        error marker is created for that item.
        """

        errmsg = "This reliable listener is not very reliable!"
        processedItems = []
        def listener(item):
            if item.information == 1:
                raise RuntimeError(errmsg)
            processedItems.append(item.information)

        proc = self.procType(store=self.store)
        listener = WorkListener(store=self.store, listener=listener)

        proc.addReliableListener(listener)

        # Make some work, step the processor, and fake the error handling
        # behavior the Scheduler actually provides.
        for i in range(3):
            TestWorkUnit(store=self.store, information=i)
            try:
                proc.step()
            except batch._ProcessingFailure:
                proc.timedEventErrorHandler(
                    (u"Oh crap, I do not have a TimedEvent, "
                     "I sure hope that never becomes a problem."),
                    failure.Failure())

        self.assertEquals(processedItems, [0, 2])

        errors = list(proc.getFailedItems())
        self.assertEquals(len(errors), 1)
        self.assertEquals(errors[0][0], listener)
        self.assertEquals(errors[0][1].information, 1)

        loggedErrors = self.flushLoggedErrors(RuntimeError)
        self.assertEquals(len(loggedErrors), 1)
        self.assertEquals(loggedErrors[0].getErrorMessage(), errmsg)


    def testMultipleListeners(self):
        """
        Test that a single batch processor with multiple listeners
        added at different times delivers each item to each listener.
        """
        processedItemsA = []
        def listenerA(item):
            processedItemsA.append(item.information)

        processedItemsB = []
        def listenerB(item):
            processedItemsB.append(item.information)

        proc = self.procType(store=self.store)

        for i in range(2):
            TestWorkUnit(store=self.store, information=i)

        firstListener = WorkListener(store=self.store, listener=listenerA)
        proc.addReliableListener(firstListener)

        for i in range(2, 4):
            TestWorkUnit(store=self.store, information=i)

        secondListener = WorkListener(store=self.store, listener=listenerB)
        proc.addReliableListener(secondListener)

        for i in range(4, 6):
            TestWorkUnit(store=self.store, information=i)

        for i in range(100):
            if not proc.step():
                break
        else:
            self.fail("Processing loop took too long")

        self.assertEquals(
            processedItemsA, [2, 3, 4, 5, 1, 0])

        self.assertEquals(
            processedItemsB, [4, 5, 3, 2, 1, 0])


    def testRepeatedAddListener(self):
        """
        Test that adding the same listener repeatedly has the same
        effect as adding it once.
        """
        proc = self.procType(store=self.store)
        listener = WorkListener(store=self.store)
        proc.addReliableListener(listener)
        proc.addReliableListener(listener)
        self.assertEquals(list(proc.getReliableListeners()), [listener])


    def testSuperfluousItemAddition(self):
        """
        Test the addItem method for work which would have been done already,
        and so for which addItem should therefore be a no-op.
        """
        processedItems = []
        def listener(item):
            processedItems.append(item.information)

        proc = self.procType(store=self.store)
        listener = WorkListener(store=self.store, listener=listener)

        # Create a couple items so there will be backwards work to do.
        one = TestWorkUnit(store=self.store, information=0)
        two = TestWorkUnit(store=self.store, information=1)

        rellist = proc.addReliableListener(listener)

        # Create a couple more items so there will be some forwards work to do.
        three = TestWorkUnit(store=self.store, information=2)
        four = TestWorkUnit(store=self.store, information=3)

        # There are only two regions at this point - work behind and work
        # ahead; no work has been done yet, so there's no region in between.
        # Add items behind and ahead of the point; these should not result in
        # any explicit tracking items, since they would have been processed in
        # due course anyway.
        rellist.addItem(two)
        rellist.addItem(three)

        for i in range(100):
            if not proc.step():
                break
        else:
            self.fail("Processing loop took too long")

        self.assertEquals(processedItems, [2, 3, 1, 0])


    def testReprocessItemAddition(self):
        """
        Test the addItem method for work which is within the bounds of work
        already done, and so which would not have been processed without the
        addItem call.
        """
        processedItems = []
        def listener(item):
            processedItems.append(item.information)

        proc = self.procType(store=self.store)
        listener = WorkListener(store=self.store, listener=listener)
        rellist = proc.addReliableListener(listener)

        one = TestWorkUnit(store=self.store, information=0)
        two = TestWorkUnit(store=self.store, information=1)
        three = TestWorkUnit(store=self.store, information=2)

        for i in range(100):
            if not proc.step():
                break
        else:
            self.fail("Processing loop took too long")

        self.assertEquals(processedItems, range(3))

        # Now that we have processed some items, re-add one of those items to
        # be re-processed and make sure it actually does get passed to the
        # listener again.
        processedItems = []

        rellist.addItem(two)

        for i in xrange(100):
            if not proc.step():
                break
        else:
            self.fail("Processing loop took too long")

        self.assertEquals(processedItems, [1])


    def test_processorStartsUnscheduled(self):
        """
        Test that when a processor is first created, it is not scheduled to
        perform any work.
        """
        proc = self.procType(store=self.store)
        self.assertIdentical(proc.scheduled, None)
        self.assertEquals(
            list(self.scheduler.scheduledTimes(proc)),
            [])


    def test_itemAddedIgnoredWithoutListeners(self):
        """
        Test that if C{itemAdded} is called while the processor is idle but
        there are no listeners, the processor does not schedule itself to do
        any work.
        """
        proc = self.procType(store=self.store)
        proc.itemAdded()
        self.assertEqual(proc.scheduled, None)
        self.assertEquals(
            list(self.scheduler.scheduledTimes(proc)),
            [])


    def test_itemAddedSchedulesProcessor(self):
        """
        Test that if C{itemAdded} is called while the processor is idle and
        there are listeners, the processor does schedules itself to do some
        work.
        """
        proc = self.procType(store=self.store)
        listener = WorkListener(store=self.store)
        proc.addReliableListener(listener)

        # Get rid of the scheduler state that addReliableListener call just
        # created.
        proc.scheduled = None
        self.scheduler.unscheduleAll(proc)

        proc.itemAdded()
        self.failIfEqual(proc.scheduled, None)
        self.assertEquals(
            list(self.scheduler.scheduledTimes(proc)),
            [proc.scheduled])


    def test_addReliableListenerSchedulesProcessor(self):
        """
        Test that if C{addReliableListener} is called while the processor is
        idle, the processor schedules itself to do some work.
        """
        proc = self.procType(store=self.store)
        listener = WorkListener(store=self.store)
        proc.addReliableListener(listener)
        self.failIfEqual(proc.scheduled, None)
        self.assertEquals(
            list(self.scheduler.scheduledTimes(proc)),
            [proc.scheduled])


    def test_itemAddedWhileScheduled(self):
        """
        Test that if C{itemAdded} is called when the processor is already
        scheduled to run, the processor remains scheduled to run at the same
        time.
        """
        proc = self.procType(store=self.store)
        listener = WorkListener(store=self.store)
        proc.addReliableListener(listener)
        when = proc.scheduled
        proc.itemAdded()
        self.assertEquals(proc.scheduled, when)
        self.assertEquals(
            list(self.scheduler.scheduledTimes(proc)),
            [proc.scheduled])


    def test_addReliableListenerWhileScheduled(self):
        """
        Test that if C{addReliableListener} is called when the processor is
        already scheduled to run, the processor remains scheduled to run at the
        same time.
        """
        proc = self.procType(store=self.store)
        listenerA = WorkListener(store=self.store)
        proc.addReliableListener(listenerA)
        when = proc.scheduled
        listenerB = WorkListener(store=self.store)
        proc.addReliableListener(listenerB)
        self.assertEquals(proc.scheduled, when)
        self.assertEquals(
            list(self.scheduler.scheduledTimes(proc)),
            [proc.scheduled])


    def test_processorIdlesWhenCaughtUp(self):
        """
        Test that the C{run} method of the processor returns C{None} when it
        has done all the work it needs to do, thus unscheduling the processor.
        """
        proc = self.procType(store=self.store)
        self.assertIdentical(proc.run(), None)



class BatchCallTestItem(item.Item):
    called = attributes.boolean(default=False)

    def callIt(self):
        self.called = True



class BrokenException(Exception):
    """
    Exception always raised by L{BrokenReliableListener.processItem}.
    """



class BatchWorkItem(item.Item):
    """
    Item class which will be delivered as work units for testing error handling
    around reliable listeners.
    """
    value = attributes.text(default=u"unprocessed")



BatchWorkSource = batch.processor(BatchWorkItem)



class BrokenReliableListener(item.Item):
    """
    A listener for batch work which always raises an exception from its
    processItem method.  Used to test that errors from processItem are properly
    handled.
    """

    anAttribute = attributes.integer()

    def processItem(self, item):
        raise BrokenException("Broken Reliable Listener is working as expected.")



class WorkingReliableListener(item.Item):
    """
    A listener for batch work which actually works.  Used to test that even if
    a broken reliable listener is around, working ones continue to receive new
    items to process.
    """

    anAttribute = attributes.integer()

    def processItem(self, item):
        item.value = u"processed"



class RemoteTestCase(unittest.TestCase):
    def test_noBatchService(self):
        """
        A L{Store} with no database directory cannot be adapted to
        L{iaxiom.IBatchService}.
        """
        st = store.Store()
        self.assertRaises(TypeError, iaxiom.IBatchService, st)
        self.assertIdentical(
            iaxiom.IBatchService(st, None), None)


    def test_subStoreNoBatchService(self):
        """
        A user L{Store} attached to a site L{Store} with no database directory
        cannot be adapted to L{iaxiom.IBatchService}.
        """
        st = store.Store(filesdir=self.mktemp())
        ss = substore.SubStore.createNew(st, ['substore']).open()
        self.assertRaises(TypeError, iaxiom.IBatchService, ss)
        self.assertIdentical(
            iaxiom.IBatchService(ss, None), None)


    def testBatchService(self):
        """
        Make sure SubStores can be adapted to L{iaxiom.IBatchService}.
        """
        dbdir = filepath.FilePath(self.mktemp())
        s = store.Store(dbdir)
        ss = substore.SubStore.createNew(s, ['substore'])
        bs = iaxiom.IBatchService(ss)
        self.failUnless(iaxiom.IBatchService.providedBy(bs))


    def testProcessLifetime(self):
        """
        Test that the batch system process can be started and stopped.
        """
        dbdir = filepath.FilePath(self.mktemp())
        s = store.Store(dbdir)
        svc = batch.BatchProcessingControllerService(s)
        svc.startService()
        return svc.stopService()


    def testCalling(self):
        """
        Test invoking a method on an item in the batch process.
        """
        dbdir = filepath.FilePath(self.mktemp())
        s = store.Store(dbdir)
        ss = substore.SubStore.createNew(s, ['substore'])
        service.IService(s).startService()
        d = iaxiom.IBatchService(ss).call(
            BatchCallTestItem(store=ss.open()).callIt)
        ss.close()
        def called(ign):
            self.assertTrue(
                ss.open().findUnique(BatchCallTestItem).called,
                "Was not called")
            ss.close()
            return service.IService(s).stopService()
        return d.addCallback(called)


    def testProcessingServiceStepsOverErrors(self):
        """
        If any processor raises an unexpected exception, the work unit which
        was being processed should be marked as having had an error and
        processing should move on to the next item.  Make sure that this
        actually happens when L{BatchProcessingService} is handling those
        errors.
        """
        BATCH_WORK_UNITS = 3

        dbdir = filepath.FilePath(self.mktemp())
        st = store.Store(dbdir)
        source = BatchWorkSource(store=st)
        for i in range(BATCH_WORK_UNITS):
            BatchWorkItem(store=st)

        source.addReliableListener(BrokenReliableListener(store=st), iaxiom.REMOTE)
        source.addReliableListener(WorkingReliableListener(store=st), iaxiom.REMOTE)

        svc = batch.BatchProcessingService(st, iaxiom.REMOTE)

        task = svc.step()

        # Loop 6 (BATCH_WORK_UNITS * 2) times - three items times two
        # listeners, it should not take any more than six iterations to
        # completely process all work.
        for i in xrange(BATCH_WORK_UNITS * 2):
            task.next()


        self.assertEquals(
            len(self.flushLoggedErrors(BrokenException)),
            BATCH_WORK_UNITS)

        self.assertEquals(
            st.query(BatchWorkItem, BatchWorkItem.value == u"processed").count(),
            BATCH_WORK_UNITS)


    def test_itemAddedStartsBatchProcess(self):
        """
        If there are remote-style listeners for an item source, C{itemAdded}
        starts the batch process.

        This is not completely correct.  There may be items to process remotely
        when the main process starts up, before any new items are added.  This
        is simpler to implement, but it shouldn't be taken as a reason not to
        implement the actually correct solution.
        """
        st = store.Store(self.mktemp())
        svc = service.IService(st)
        svc.startService()
        self.addCleanup(svc.stopService)

        batchService = iaxiom.IBatchService(st)

        procType = batch.processor(TestWorkUnit)
        proc = procType(store=st)
        listener = WorkListener(store=st)
        proc.addReliableListener(listener, style=iaxiom.REMOTE)

        # Sanity check: addReliableListener should eventually also trigger a
        # batch process start if necessary.  But we don't want to test that case
        # here, so make sure it's not happening.
        self.assertEquals(batchService.batchController.mode, 'stopped')

        # Now trigger it to start.
        proc.itemAdded()

        # It probably won't be ready by now, but who knows.
        self.assertIn(batchService.batchController.mode, ('starting', 'ready'))


    def test_itemAddedBeforeStarted(self):
        """
        If C{itemAdded} is called before the batch service is started, the batch
        process is not started.
        """
        st = store.Store(self.mktemp())

        procType = batch.processor(TestWorkUnit)
        proc = procType(store=st)
        listener = WorkListener(store=st)
        proc.addReliableListener(listener, style=iaxiom.REMOTE)

        proc.itemAdded()

        # When the service later starts, the batch service needn't start its
        # process.  Not that this would be bad.  Feel free to reverse this
        # behavior if you really want.
        svc = service.IService(st)
        svc.startService()
        self.addCleanup(svc.stopService)

        batchService = iaxiom.IBatchService(st)
        self.assertEquals(batchService.batchController.mode, 'stopped')


    def test_itemAddedWithoutBatchService(self):
        """
        If the store has no batch service, C{itemAdded} doesn't start the batch
        process and also doesn't raise an exception.
        """
        # An in-memory store can't have a batch service.
        st = store.Store()
        svc = service.IService(st)
        svc.startService()
        self.addCleanup(svc.stopService)

        procType = batch.processor(TestWorkUnit)
        proc = procType(store=st)
        listener = WorkListener(store=st)
        proc.addReliableListener(listener, style=iaxiom.REMOTE)

        proc.itemAdded()

        # And still there should be no batch service at all.
        self.assertIdentical(iaxiom.IBatchService(st, None), None)


    def test_subStoreBatchServiceStart(self):
        """
        The substore implementation of L{IBatchService.start} starts the batch
        process.
        """
        st = store.Store(self.mktemp())
        svc = service.IService(st)
        svc.startService()
        self.addCleanup(svc.stopService)

        ss = substore.SubStore.createNew(st, ['substore']).open()
        iaxiom.IBatchService(ss).start()

        batchService = iaxiom.IBatchService(st)
        self.assertIn(batchService.batchController.mode, ('starting', 'ready'))



# Copright 2008 Divmod, Inc.  See LICENSE file for details.

from zope.interface import Interface, implements

from twisted.trial import unittest

from axiom import dependency
from axiom.store import Store
from axiom.substore import SubStore
from axiom.item import Item
from axiom.errors import UnsatisfiedRequirement
from axiom.attributes import text, integer, reference, inmemory


class IElectricityGrid(Interface):
    """
    An interface representing something likely to be present in the site store.
    As opposed to the other examples below, present in a hypothetical kitchen,
    it is something managed for lots of different people.
    """

    def draw(watts):
        """
        Draw some number of watts from this power grid.

        @return: a constant, one of L{REAL_POWER} or L{FAKE_POWER}.
        """


FAKE_POWER = 'fake power'
REAL_POWER = 'real power'

class NullGrid(object):
    """
    This is a null electricity grid.  It is provided as a default grid in the
    case where a site store is not present.
    """
    implements(IElectricityGrid)

    def __init__(self, siteStore):
        """
        Create a null grid with a reference to the site store.
        """
        self.siteStore = siteStore


    def draw(self, watts):
        """
        Draw some watts from the null power grid.  For simplicity of examples
        below, this works.  Not in real life, though.  In a more realistic
        example, this might do something temporary to work around the site
        misconfiguration, and warn an administrator that someone was getting
        power out of thin air.  Or, depending on the application, we might
        raise an exception to prevent this operation from succeeding.
        """
        return FAKE_POWER


class RealGrid(Item):
    """
    A power grid for the power utility; this is an item which should be
    installed on a site store.
    """
    implements(IElectricityGrid)

    powerupInterfaces = (IElectricityGrid,)

    totalWattage = integer(default=10000000,
                           doc="""
                           Total wattage of the entire electricity grid.  (This
                           is currently a dummy attribute.)
                           """)

    def draw(self, watts):
        """
        Draw some real power from the real power grid.  This is the way that
        the site should probably be working.
        """
        return REAL_POWER



def noGrid(siteStore):
    """
    No power grid was available.  Raise an exception.
    """
    raise RuntimeError("No power grid available.")



class IronLung(Item):
    """
    This item is super serious business!  It has to draw real power from the
    real power grid; it won't be satisfied with fake power; too risky for its
    life-critical operation.  So it doesn't specify a placeholder default grid.

    @ivar grid: a read-only reference to an L{IElectricityGrid} provider,
    resolved via the site store this L{IronLung} is in.
    """

    wattsPerPump = integer(default=100, allowNone=False,
                           doc="""
                           The number of watts to draw from L{self.grid} when
                           L{IronLung.pump} is called.
                           """)

    grid = dependency.requiresFromSite(IElectricityGrid)

    def pump(self):
        """
        Attempting to pump the iron lung by talking to the power grid.
        """
        return self.grid.draw(self.wattsPerPump)



class SpecifiedBadDefaults(Item):
    """
    Depends on a power grid, but specifies defaults for that dependency that
    should never be invoked.  This item can't retrieve a grid.

    @ivar grid: Retrieving this attribute should never work.  It should raise
    L{RuntimeError}.
    """
    dummy = integer(doc="""
    Dummy attribute required by axiom for Item class definition.
    """)

    grid = dependency.requiresFromSite(IElectricityGrid, noGrid, noGrid)

    def pump(self):
        """
        Attempting to pump the iron lung by talking to the power grid.
        """
        return self.grid.draw(100)


class Kitchen(Item):
    name = text()

class PowerStrip(Item):
    """
    A simulated collection of power points.  This is where L{IAppliance}
    providers get their power from.

    @ivar grid: A read-only reference to an L{IElectricityGrid} provider.  This
    may be a powerup provided by the site store or a L{NullGrid} if no powerup
    is installed.
    """
    voltage = integer()
    grid = dependency.requiresFromSite(IElectricityGrid, NullGrid, NullGrid)

    def setForUSElectricity(self):
        if not self.voltage:
            self.voltage = 110
        else:
            raise RuntimeError("Oops! power strip already set up")

    def draw(self, watts):
        """
        Draw the given amount of power from this strip's electricity grid.

        @param watts: The number of watts to draw.

        @type watts: L{int}
        """
        return self.grid.draw(watts)


class PowerPlant(Item):
    """
    This is an item which supplies the grid with power.  It lives in the site
    store.

    @ivar grid: a read-only reference to an L{IElectricityGrid} powerup on the
    site store, or a L{NullGrid} if none is installed.  If this item is present
    in a user store, retrieving this will raise a L{RuntimeError}.
    """

    wattage = integer(default=1000, allowNone=False,
                      doc="""
                      The amount of power the grid will be supplied with.
                      Currently a dummy attribute required by axiom for item
                      definition.
                      """)
    grid = dependency.requiresFromSite(IElectricityGrid, noGrid, NullGrid)



class IAppliance(Interface):
    pass

class IBreadConsumer(Interface):
    pass

class Breadbox(Item):
    slices = integer(default=100)

    def dispenseBread(self, amt):
        self.slices -= amt

class Toaster(Item):
    implements(IBreadConsumer)
    powerupInterfaces = (IAppliance, IBreadConsumer)

    powerStrip = dependency.dependsOn(PowerStrip,
                                      lambda ps: ps.setForUSElectricity(),
                                      doc="the power source for this toaster")
    description = text()
    breadFactory = dependency.dependsOn(
        Breadbox,
        doc="the thing we get bread input from",
        whenDeleted=reference.CASCADE)

    callback = inmemory()

    def activate(self):
        self.callback = None

    def installed(self):
        if self.callback is not None:
            self.callback("installed")

    def uninstalled(self):
        if self.callback is not None:
            self.callback("uninstalled")

    def toast(self):
        self.powerStrip.draw(100)
        self.breadFactory.dispenseBread(2)

def powerstripSetup(ps):
    ps.setForUSElectricity()
class Blender(Item):
    powerStrip = dependency.dependsOn(PowerStrip,
                                      powerstripSetup)
    description = text()

    def __getPowerupInterfaces__(self, powerups):
        yield (IAppliance, 0)

class IceCrusher(Item):
    blender = dependency.dependsOn(Blender)

class Blender2(Item):
    powerStrip = reference()

class DependencyTest(unittest.TestCase):
    def setUp(self):
        self.store = Store()

    def test_dependsOn(self):
        """
        Ensure that classes with dependsOn attributes set up the dependency map
        properly.
        """
        foo = Blender(store=self.store)
        depBlob = dependency._globalDependencyMap.get(Blender, None)[0]
        self.assertEqual(depBlob[0], PowerStrip)
        self.assertEqual(depBlob[1], powerstripSetup)
        self.assertEqual(depBlob[2], Blender.__dict__['powerStrip'])

    def test_classDependsOn(self):
        """
        Ensure that classDependsOn sets up the dependency map properly.
        """
        dependency.classDependsOn(Blender2, PowerStrip, powerstripSetup,
                                  Blender2.__dict__['powerStrip'])
        depBlob = dependency._globalDependencyMap.get(Blender2, None)[0]
        self.assertEqual(depBlob[0], PowerStrip)
        self.assertEqual(depBlob[1], powerstripSetup)
        self.assertEqual(depBlob[2], Blender2.__dict__['powerStrip'])

    def test_basicInstall(self):
        """
        If a Toaster gets installed in a Kitchen, make sure that the
        required dependencies get instantiated and installed too.
        """
        foo = Kitchen(store=self.store)
        e = Toaster(store=self.store)
        self.assertEquals(e.powerStrip, None)
        dependency.installOn(e, foo)
        e.toast()
        ps = self.store.findUnique(PowerStrip, default=None)
        bb = self.store.findUnique(Breadbox, default=None)
        self.failIfIdentical(ps, None)
        self.failIfIdentical(bb, None)
        self.assertEquals(e.powerStrip, ps)
        self.assertEquals(ps.voltage, 110)
        self.assertEquals(e.breadFactory, bb)
        self.assertEquals(set(dependency.installedRequirements(e, foo)),
                          set([ps, bb]))
        self.assertEquals(list(dependency.installedDependents(ps, foo)), [e])

    def test_basicUninstall(self):
        """
        Ensure that uninstallation removes the adapter from the former
        install target and all orphaned dependencies.
        """
        foo = Kitchen(store=self.store)
        e = Toaster(store=self.store)
        dependency.installOn(e, foo)
        dependency.uninstallFrom(e, foo)
        self.assertEqual(dependency.installedOn(e), None)
        self.assertEqual(dependency.installedOn(e.powerStrip), None)

    def test_wrongUninstall(self):
        """
        Ensure that attempting to uninstall an item that something
        else depends on fails.
        """
        foo = Kitchen(store=self.store)
        e = Toaster(store=self.store)
        dependency.installOn(e, foo)

        ps = self.store.findUnique(PowerStrip)
        self.failUnlessRaises(dependency.DependencyError,
                              dependency.uninstallFrom, ps, foo)

    def test_properOrphaning(self):
        """
        If two installed items both depend on a third, it should be
        removed as soon as both installed items are removed, but no
        sooner.
        """

        foo = Kitchen(store=self.store)
        e = Toaster(store=self.store)
        dependency.installOn(e, foo)
        ps = self.store.findUnique(PowerStrip)
        bb = self.store.findUnique(Breadbox)
        f = Blender(store=self.store)
        dependency.installOn(f, foo)

        self.assertEquals(list(self.store.query(PowerStrip)), [ps])
        #XXX does ordering matter?
        self.assertEquals(set(dependency.installedDependents(ps, foo)),
                          set([e, f]))
        self.assertEquals(set(dependency.installedRequirements(e, foo)),
                          set([bb, ps]))
        self.assertEquals(list(dependency.installedRequirements(f, foo)),
                          [ps])

        dependency.uninstallFrom(e, foo)
        self.assertEquals(dependency.installedOn(ps), foo)

        dependency.uninstallFrom(f, foo)
        self.assertEquals(dependency.installedOn(ps), None)

    def test_installedUniqueRequirements(self):
        """
        Ensure that installedUniqueRequirements lists only powerups depended on
        by exactly one installed powerup.
        """
        foo = Kitchen(store=self.store)
        e = Toaster(store=self.store)
        dependency.installOn(e, foo)
        ps = self.store.findUnique(PowerStrip)
        bb = self.store.findUnique(Breadbox)
        f = Blender(store=self.store)
        dependency.installOn(f, foo)

        self.assertEquals(list(dependency.installedUniqueRequirements(e, foo)),
                          [bb])

    def test_customizerCalledOnce(self):
        """
        The item customizer defined for a dependsOn attribute should
        only be called if an item is created implicitly to satisfy the
        dependency.
        """
        foo = Kitchen(store=self.store)
        ps = PowerStrip(store=self.store)
        dependency.installOn(ps, foo)
        ps.voltage = 115
        e = Toaster(store=self.store)
        dependency.installOn(e, foo)
        self.assertEqual(ps.voltage, 115)

    def test_explicitInstall(self):
        """
        If an item is explicitly installed, it should not be
        implicitly uninstalled. Also, dependsOn attributes should be
        filled in properly even if a dependent item is not installed
        automatically.
        """
        foo = Kitchen(store=self.store)
        ps = PowerStrip(store=self.store)
        dependency.installOn(ps, foo)
        e = Toaster(store=self.store)
        dependency.installOn(e, foo)
        self.assertEqual(e.powerStrip, ps)
        dependency.uninstallFrom(e, foo)
        self.assertEquals(dependency.installedOn(ps), foo)

    def test_doubleInstall(self):
        """
        Make sure that installing two instances of a class on the same
        target fails, if something depends on that class, and succeeds
        otherwise.
        """
        foo = Kitchen(store=self.store)
        e = Toaster(store=self.store)
        dependency.installOn(e, foo)
        ps = PowerStrip(store=self.store)
        self.failUnlessRaises(dependency.DependencyError,
                              dependency.installOn, ps, foo)
        e2 = Toaster(store=self.store)
        dependency.installOn(e2, foo)


    def test_recursiveInstall(self):
        """
        Installing an item should install all of its dependencies, and
        all of its dependencies, and so forth.
        """
        foo = Kitchen(store=self.store)
        ic = IceCrusher(store=self.store)
        dependency.installOn(ic, foo)
        blender = self.store.findUnique(Blender)
        ps = self.store.findUnique(PowerStrip)

        self.assertEquals(dependency.installedOn(blender), foo)
        self.assertEquals(dependency.installedOn(ps), foo)
        self.assertEquals(list(dependency.installedRequirements(ic, foo)),
                          [blender])

    def test_recursiveUninstall(self):
        """
        Removal of items should recursively remove orphaned
        dependencies.
        """
        foo = Kitchen(store=self.store)
        ic = IceCrusher(store=self.store)
        dependency.installOn(ic, foo)
        blender = self.store.findUnique(Blender)
        ps = self.store.findUnique(PowerStrip)

        dependency.uninstallFrom(ic, foo)

        self.failIf(dependency.installedOn(blender))
        self.failIf(dependency.installedOn(ps))
        self.failIf(dependency.installedOn(ic))

    def test_wrongDependsOn(self):
        """
        dependsOn should raise an error if used outside a class definition.
        """
        self.assertRaises(TypeError, dependency.dependsOn, Toaster)

    def test_referenceArgsPassthrough(self):
        """
        dependsOn should accept (most of) attributes.reference's args.
        """

        self.failUnless("power source" in Toaster.powerStrip.doc)
        self.assertEquals(Toaster.breadFactory.whenDeleted, reference.CASCADE)

    def test_powerupInterfaces(self):
        """
        Make sure interfaces are powered up and down properly.
        """

        foo = Kitchen(store=self.store)
        e = Toaster(store=self.store)
        f = Blender(store=self.store)
        dependency.installOn(e, foo)
        dependency.installOn(f, foo)
        self.assertEquals(IAppliance(foo), e)
        self.assertEquals(IBreadConsumer(foo), e)
        dependency.uninstallFrom(e, foo)
        self.assertEquals(IAppliance(foo), f)
        dependency.uninstallFrom(f, foo)
        self.assertRaises(TypeError, IAppliance, foo)


    def test_callbacks(self):
        """
        'installed' and 'uninstalled' callbacks should fire on
        install/uninstall.
        """
        foo = Kitchen(store=self.store)
        e = Toaster(store=self.store)
        self.installCallbackCalled = False
        e.callback = lambda _: setattr(self, 'installCallbackCalled', True)
        dependency.installOn(e, foo)
        self.failUnless(self.installCallbackCalled)
        self.uninstallCallbackCalled = False
        e.callback = lambda _: setattr(self, 'uninstallCallbackCalled', True)
        dependency.uninstallFrom(e, foo)
        self.failUnless(self.uninstallCallbackCalled)


    def test_onlyInstallPowerups(self):
        """
        Make sure onlyInstallPowerups doesn't load dependencies or prohibit
        multiple calls.
        """
        foo = Kitchen(store=self.store)
        e = Toaster(store=self.store)
        f = Toaster(store=self.store)
        dependency.onlyInstallPowerups(e, foo)
        dependency.onlyInstallPowerups(f, foo)
        self.assertEquals(list(foo.powerupsFor(IBreadConsumer)), [e, f])
        self.assertEquals(list(self.store.query(
                    dependency._DependencyConnector)), [])


class RequireFromSiteTests(unittest.TestCase):
    """
    L{axiom.dependency.requiresFromSite} should allow items in either a user or
    site store to depend on powerups in the site store.
    """

    def setUp(self):
        """
        Create a L{Store} to be used as the site store for these tests.
        """
        self.store = Store()


    def test_requiresFromSite(self):
        """
        The value of a L{axiom.dependency.requiresFromSite} descriptor ought to
        be the powerup on the site for the instance it describes.
        """
        dependency.installOn(RealGrid(store=self.store), self.store)
        substore = SubStore.createNew(self.store, ['sub']).open()
        self.assertEquals(PowerStrip(store=substore).draw(1), REAL_POWER)


    def test_requiresFromSiteDefault(self):
        """
        The value of a L{axiom.dependency.requiresFromSite} descriptor on an
        item in a user store ought to be the result of invoking its default
        factory parameter.
        """
        substore = SubStore.createNew(self.store, ['sub']).open()
        ps = PowerStrip(store=substore)
        self.assertEquals(ps.draw(1), FAKE_POWER)
        self.assertEquals(ps.grid.siteStore, self.store)


    def test_requiresFromSiteInSiteStore(self):
        """
        L{axiom.dependency.requiresFromSite} should use the
        C{siteDefaultFactory} rather than the C{defaultFactory} to satisfy the
        dependency for items stored in a site store.  It should use this
        default whether or not any item which could satisfy the requirement is
        installed on the site store.

        This behavior is important because some powerup interfaces are provided
        for site and user stores with radically different behaviors; for
        example, the substore implementation of L{IScheduler} depends on the
        site implementation of L{IScheduler}; if a user's substore were opened
        accidentally as a site store (i.e. with no parent) then the failure of
        the scheduler API should be obvious and immediate so that it can
        compensate; it should not result in an infinite recursion as the
        scheduler is looking for its parent.

        Items which wish to be stored in a site store and also depend on items
        in the site store can specifically adapt to the appropriate interface
        in the C{siteDefaultFactory} supplied to
        L{dependency.requiresFromSite}.
        """
        plant = PowerPlant(store=self.store)
        self.assertEquals(plant.grid.siteStore, self.store)
        self.assertEquals(plant.grid.draw(100), FAKE_POWER)
        dependency.installOn(RealGrid(store=self.store), self.store)
        self.assertEquals(plant.grid.siteStore, self.store)
        self.assertEquals(plant.grid.draw(100), FAKE_POWER)


    def test_requiresFromSiteNoDefault(self):
        """
        The default function shouldn't be needed or invoked if its value isn't
        going to be used.
        """
        dependency.installOn(RealGrid(store=self.store), self.store)
        substore = SubStore.createNew(self.store, ['sub']).open()
        self.assertEquals(SpecifiedBadDefaults(store=substore).pump(),
                          REAL_POWER)


    def test_requiresFromSiteUnspecifiedException(self):
        """
        If a default factory function isn't supplied, an
        L{UnsatisfiedRequirement}, which should be a subtype of
        L{AttributeError}, should be raised when the descriptor is retrieved.
        """
        lung = IronLung(store=self.store)
        siteLung = IronLung(
            store=SubStore.createNew(self.store, ['sub']).open())
        self.assertRaises(UnsatisfiedRequirement, lambda : lung.grid)
        self.assertRaises(UnsatisfiedRequirement, lambda : siteLung.grid)
        default = object()
        self.assertIdentical(getattr(lung, 'grid', default), default)

